@@ -40,7 +40,6 @@
 #include <minmax.h>
 #include <errno.h>
 #include <suffix_number.h>
-#include <graphics.h>
 #include <dprintf.h>
 
 #include <syslinux/align.h>
@@ -48,6 +47,7 @@
 #include <syslinux/bootrm.h>
 #include <syslinux/movebits.h>
 #include <syslinux/firmware.h>
+#include <syslinux/video.h>
 
 #define BOOT_MAGIC 0xAA55
 #define LINUX_MAGIC ('H' + ('d' << 8) + ('r' << 16) + ('S' << 24))
@@ -124,14 +124,39 @@ static int map_initramfs(struct syslinux_movelist **fraglist,
     return 0;
 }
 
+static size_t calc_cmdline_offset(const struct syslinux_memmap *mmap,
+				  const struct linux_header *hdr,
+				  size_t cmdline_size, addr_t base,
+				  addr_t start)
+{
+    size_t max_offset;
+
+    if (hdr->version >= 0x0202 && (hdr->loadflags & LOAD_HIGH))
+	max_offset = 0x10000;
+    else
+	max_offset = 0xfff0 - cmdline_size;
+
+    if (!syslinux_memmap_highest(mmap, SMT_FREE, &start,
+				 cmdline_size, 0xa0000, 16) ||
+	!syslinux_memmap_highest(mmap, SMT_TERMINAL, &start,
+				 cmdline_size, 0xa0000, 16)) {
+	
+
+	return min(start - base, max_offset) & ~15;
+    }
+
+    dprintf("Unable to find lowmem for cmdline\n");
+    return (0x9ff0 - cmdline_size) & ~15; /* Legacy value: pure hope... */
+}
+
 int bios_boot_linux(void *kernel_buf, size_t kernel_size,
 		    struct initramfs *initramfs,
 		    struct setup_data *setup_data,
 		    char *cmdline)
 {
     struct linux_header hdr, *whdr;
-    size_t real_mode_size, prot_mode_size;
-    addr_t real_mode_base, prot_mode_base;
+    size_t real_mode_size, prot_mode_size, base;
+    addr_t real_mode_base, prot_mode_base, prot_mode_max;
     addr_t irf_size;
     size_t cmdline_size, cmdline_offset;
     struct setup_data *sdp;
@@ -139,7 +164,6 @@ int bios_boot_linux(void *kernel_buf, size_t kernel_size,
     struct syslinux_movelist *fraglist = NULL;
     struct syslinux_memmap *mmap = NULL;
     struct syslinux_memmap *amap = NULL;
-    bool ok;
     uint32_t memlimit = 0;
     uint16_t video_mode = 0;
     const char *arg;
@@ -147,8 +171,10 @@ int bios_boot_linux(void *kernel_buf, size_t kernel_size,
     cmdline_size = strlen(cmdline) + 1;
 
     errno = EINVAL;
-    if (kernel_size < 2 * 512)
+    if (kernel_size < 2 * 512) {
+	dprintf("Kernel size too small\n");
 	goto bail;
+    }
 
     /* Look for specific command-line arguments we care about */
     if ((arg = find_argument(cmdline, "mem=")))
@@ -179,8 +205,10 @@ int bios_boot_linux(void *kernel_buf, size_t kernel_size,
     memcpy(&hdr, kernel_buf, sizeof hdr);
     whdr = (struct linux_header *)kernel_buf;
 
-    if (hdr.boot_flag != BOOT_MAGIC)
+    if (hdr.boot_flag != BOOT_MAGIC) {
+	dprintf("Invalid boot magic\n");
 	goto bail;
+    }
 
     if (hdr.header != LINUX_MAGIC) {
 	hdr.version = 0x0100;	/* Very old kernel */
@@ -192,7 +220,7 @@ int bios_boot_linux(void *kernel_buf, size_t kernel_size,
     if (!hdr.setup_sects)
 	hdr.setup_sects = 4;
 
-    if (hdr.version < 0x0203)
+    if (hdr.version < 0x0203 || !hdr.initrd_addr_max)
 	hdr.initrd_addr_max = 0x37ffffff;
 
     if (!memlimit && memlimit - 1 > hdr.initrd_addr_max)
@@ -209,16 +237,25 @@ int bios_boot_linux(void *kernel_buf, size_t kernel_size,
 	cmdline[cmdline_size - 1] = '\0';
     }
 
-    if (hdr.version < 0x0202 || !(hdr.loadflags & 0x01))
-	cmdline_offset = (0x9ff0 - cmdline_size) & ~15;
-    else
-	cmdline_offset = 0x10000;
-
     real_mode_size = (hdr.setup_sects + 1) << 9;
     real_mode_base = (hdr.loadflags & LOAD_HIGH) ? 0x10000 : 0x90000;
     prot_mode_base = (hdr.loadflags & LOAD_HIGH) ? 0x100000 : 0x10000;
+    prot_mode_max  = (hdr.loadflags & LOAD_HIGH) ? (addr_t)-1 : 0x8ffff;
     prot_mode_size = kernel_size - real_mode_size;
 
+    /* Get the memory map */
+    mmap = syslinux_memory_map();	/* Memory map for shuffle_boot */
+    amap = syslinux_dup_memmap(mmap);	/* Keep track of available memory */
+    if (!mmap || !amap) {
+	errno = ENOMEM;
+	goto bail;
+    }
+
+    cmdline_offset = calc_cmdline_offset(mmap, &hdr, cmdline_size,
+					 real_mode_base,
+					 real_mode_base + real_mode_size);
+    dprintf("cmdline_offset at 0x%x\n", real_mode_base + cmdline_offset);
+
     if (hdr.version < 0x020a) {
 	/*
 	 * The 3* here is a total fudge factor... it's supposed to
@@ -230,14 +267,18 @@ int bios_boot_linux(void *kernel_buf, size_t kernel_size,
 	hdr.init_size = 3 * prot_mode_size;
     }
 
-    if (!(hdr.loadflags & LOAD_HIGH) && prot_mode_size > 512 * 1024)
-	goto bail;		/* Kernel cannot be loaded low */
+    if (!(hdr.loadflags & LOAD_HIGH) && prot_mode_size > 512 * 1024) {
+	dprintf("Kernel cannot be loaded low\n");
+	goto bail;
+    }
 
     /* Get the size of the initramfs, if there is one */
     irf_size = initramfs_size(initramfs);
 
-    if (irf_size && hdr.version < 0x0200)
-	goto bail;		/* initrd/initramfs not supported */
+    if (irf_size && hdr.version < 0x0200) {
+	dprintf("Initrd specified but not supported by kernel\n");
+	goto bail;
+    }
 
     if (hdr.version >= 0x0200) {
 	whdr->type_of_loader = 0x30;	/* SYSLINUX unknown module */
@@ -247,14 +288,6 @@ int bios_boot_linux(void *kernel_buf, size_t kernel_size,
 	}
     }
 
-    /* Get the memory map */
-    mmap = syslinux_memory_map();	/* Memory map for shuffle_boot */
-    amap = syslinux_dup_memmap(mmap);	/* Keep track of available memory */
-    if (!mmap || !amap) {
-	errno = ENOMEM;
-	goto bail;
-    }
-
     dprintf("Initial memory map:\n");
     syslinux_dump_memmap(mmap);
 
@@ -270,77 +303,33 @@ int bios_boot_linux(void *kernel_buf, size_t kernel_size,
 
     /* Place the kernel in memory */
 
-    /* First, find a suitable place for the protected-mode code */
+    /*
+     * First, find a suitable place for the protected-mode code.  If
+     * the kernel image is not relocatable, just worry if it fits (it
+     * might not even be a Linux image, after all, and for !LOAD_HIGH
+     * we end up decompressing into a different location anyway), but
+     * if it is, make sure everything fits.
+     */
+    base = prot_mode_base;
     if (prot_mode_size &&
-	syslinux_memmap_type(amap, prot_mode_base, prot_mode_size)
-	!= SMT_FREE) {
-	const struct syslinux_memmap *mp;
-	if (!hdr.relocatable_kernel)
-	    goto bail;		/* Can't relocate - no hope */
-
-	ok = false;
-	for (mp = amap; mp; mp = mp->next) {
-	    addr_t start, end;
-	    start = mp->start;
-	    end = mp->next->start;
-
-	    if (mp->type != SMT_FREE)
-		continue;
-
-	    if (end <= prot_mode_base)
-		continue;	/* Only relocate upwards */
-
-	    if (start <= prot_mode_base)
-		start = prot_mode_base;
-
-	    start = ALIGN_UP(start, hdr.kernel_alignment);
-	    if (start >= end)
-		continue;
-
-	    if (end - start >= hdr.init_size) {
-		whdr->code32_start += start - prot_mode_base;
-		prot_mode_base = start;
-		ok = true;
-		break;
-	    }
-	}
-
-	if (!ok)
-	    goto bail;
+	syslinux_memmap_find(amap, &base,
+			     hdr.relocatable_kernel ?
+			     hdr.init_size : prot_mode_size,
+			     hdr.relocatable_kernel, hdr.kernel_alignment,
+			     prot_mode_base, prot_mode_max,
+			     prot_mode_base, prot_mode_max)) {
+	dprintf("Could not find location for protected-mode code\n");
+	goto bail;
     }
 
+    whdr->code32_start += base - prot_mode_base;
+
     /* Real mode code */
-    if (syslinux_memmap_type(amap, real_mode_base,
-			     cmdline_offset + cmdline_size) != SMT_FREE) {
-	const struct syslinux_memmap *mp;
-
-	ok = false;
-	for (mp = amap; mp; mp = mp->next) {
-	    addr_t start, end;
-	    start = mp->start;
-	    end = mp->next->start;
-
-	    if (mp->type != SMT_FREE)
-		continue;
-
-	    if (start < real_mode_base)
-		start = real_mode_base;	/* Lowest address we'll use */
-	    if (end > 640 * 1024)
-		end = 640 * 1024;
-
-	    start = ALIGN_UP(start, 16);
-	    if (start > 0x90000 || start >= end)
-		continue;
-
-	    if (end - start >= cmdline_offset + cmdline_size) {
-		real_mode_base = start;
-		ok = true;
-		break;
-	    }
-	}
-
-	if (!ok)
-	    goto bail;
+    if (syslinux_memmap_find(amap, &real_mode_base,
+			     cmdline_offset + cmdline_size, true, 16,
+			     real_mode_base, 0x90000, 0, 640*1024)) {
+	dprintf("Could not find location for real-mode code\n");
+	goto bail;
     }
 
     if (syslinux_add_movelist(&fraglist, real_mode_base, (addr_t) kernel_buf,
@@ -408,8 +397,10 @@ int bios_boot_linux(void *kernel_buf, size_t kernel_size,
 		    best_addr = (adj_end - irf_size) & ~align_mask;
 	    }
 
-	    if (!best_addr)
-		goto bail;	/* Insufficient memory for initramfs */
+	    if (!best_addr) {
+		dprintf("Insufficient memory for initramfs\n");
+		goto bail;
+	    }
 
 	    whdr->ramdisk_image = best_addr;
 	    whdr->ramdisk_size = irf_size;
@@ -504,6 +495,7 @@ int bios_boot_linux(void *kernel_buf, size_t kernel_size,
     }
 
     syslinux_shuffle_boot_rm(fraglist, mmap, 0, &regs);
+    dprintf("shuffle_boot_rm failed\n");
 
 bail:
     syslinux_free_movelist(fraglist);
@@ -517,11 +509,10 @@ int syslinux_boot_linux(void *kernel_buf, size_t kernel_size,
 			struct setup_data *setup_data,
 			char *cmdline)
 {
-    if (!firmware->boot_linux) {
-	printf("No linux boot function registered for firmware\n");
-	return -1;
-    }
+    if (firmware->boot_linux)
+	return firmware->boot_linux(kernel_buf, kernel_size, initramfs,
+				    setup_data, cmdline);
 
-    return firmware->boot_linux(kernel_buf, kernel_size, initramfs,
-				setup_data, cmdline);
+    return bios_boot_linux(kernel_buf, kernel_size, initramfs,
+			   setup_data, cmdline);
 }
