@@ -43,19 +43,12 @@
 #include <dprintf.h>
 #include <syslinux/movebits.h>
 #include <klibc/compiler.h>
+#include <syslinux/boot.h>
 
 struct shuffle_descriptor {
     uint32_t dst, src, len;
 };
 
-static int shuffler_size;
-
-static void __constructor __syslinux_get_shuffer_size(void)
-{
-    /* +15 padding is to guarantee alignment */
-    shuffler_size = __bcopyxx_len + 15;
-}
-
 /*
  * Allocate descriptor memory in these chunks; if this is large we may
  * waste memory, if it is small we may get slow convergence.
@@ -76,7 +69,11 @@ int syslinux_do_shuffle(struct syslinux_movelist *fraglist,
     int need_ptrs;
     addr_t desczone, descfree, descaddr;
     int nmoves, nzero;
-    com32sys_t ireg;
+
+#ifndef __FIRMWARE_BIOS__
+    errno = ENOSYS;
+    return -1;			/* Not supported at this time*/
+#endif
 
     descaddr = 0;
     dp = dbuf = NULL;
@@ -120,7 +117,8 @@ int syslinux_do_shuffle(struct syslinux_movelist *fraglist,
 	   descriptor, plus the shuffler safe area. */
 	addr_t descmem = desc_blocks *
 	    sizeof(struct shuffle_descriptor) * DESC_BLOCK_SIZE
-	    + sizeof(struct shuffle_descriptor) + shuffler_size;
+	    + sizeof(struct shuffle_descriptor)
+	    + syslinux_shuffler_size();
 
 	descaddr = (desczone + descfree - descmem) & ~3;
 
@@ -221,13 +219,8 @@ bail:
 	return rv;
 
     /* Actually do it... */
-    memset(&ireg, 0, sizeof ireg);
-    ireg.edi.l = descaddr;
-    ireg.esi.l = (addr_t) dbuf;
-    ireg.ecx.l = (addr_t) dp - (addr_t) dbuf;
-    ireg.edx.w[0] = bootflags;
-    ireg.eax.w[0] = 0x0024;
-    __intcall(0x22, &ireg, NULL);
+    bios_do_shuffle_and_boot(bootflags, descaddr, dbuf,
+			     (size_t)dp - (size_t)dbuf);
 
     return -1;			/* Shouldn't have returned! */
 }
