@@ -1,15 +1,19 @@
 #include <syslinux/firmware.h>
+#include <syslinux/memscan.h>
 #include <core.h>
 #include "pxe.h"
 #include <net.h>
 #include <minmax.h>
 #include <bios.h>
+#include <dprintf.h>
 
 static uint16_t real_base_mem;	   /* Amount of DOS memory after freeing */
 
 static bool has_gpxe;
 static uint32_t gpxe_funcs;
 
+far_ptr_t StrucPtr;
+
 /*
  * Validity check on possible !PXE structure in buf
  * return 1 for success, 0 for failure.
@@ -88,6 +92,29 @@ static const struct pxenv_t *memory_scan_for_pxenv_struct(void)
     return memory_scan(0x10000, is_pxenv);
 }
 
+static int pxelinux_scan_memory(scan_memory_callback_t callback, void *data)
+{
+    addr_t start, size;
+    int rv = 0;
+
+    if (KeepPXE)
+	return 0;
+
+    /*
+     * If we are planning on calling unload_pxe() and unmapping the PXE
+     * region before we transfer control away from PXELINUX we can mark
+     * that region as SMT_TERMINAL to indicate that the region will
+     * become free at some point in the future.
+     */
+    start = bios_fbm() << 10;
+    size = (real_base_mem - bios_fbm()) << 10;
+    dprintf("Marking PXE region 0x%x - 0x%x as SMT_TERMINAL\n",
+	start, start + size);
+
+    callback(data, start, size, SMT_TERMINAL);
+    return rv;
+}
+
 /*
  * Find the !PXE structure; we search for the following, in order:
  *
@@ -138,30 +165,38 @@ int pxe_init(bool quiet)
     regs.eax.w[0] = 0x5650;
     call16(pxe_int1a, &regs, &regs);
     if (!(regs.eflags.l & EFLAGS_CF) && (regs.eax.w[0] == 0x564e)) {
-	pxenv = MK_PTR(regs.es, regs.ebx.w[0]);
+	off = regs.ebx.w[0];
+	seg = regs.es;
+	pxenv = MK_PTR(seg, off);
         if (is_pxenv(pxenv))
             goto have_pxenv;
     }
 
     /* Plan D: !PXE memory scan */
     plan++;
-    if ((pxe = memory_scan_for_pxe_struct()))
+    if ((pxe = memory_scan_for_pxe_struct())) {
+	off = OFFS(pxe);
+	seg = SEG(pxe);
         goto have_pxe;
+    }
 
     /* Plan E: PXENV+ memory scan */
     plan++;
-    if ((pxenv = memory_scan_for_pxenv_struct()))
+    if ((pxenv = memory_scan_for_pxenv_struct())) {
+	off = OFFS(pxenv);
+	seg = SEG(pxenv);
         goto have_pxenv;
+    }
 
     /* Found nothing at all !! */
     if (!quiet)
-	printf("No !PXE or PXENV+ API found; we're dead...\n");
+	ddprintf("No !PXE or PXENV+ API found; we're dead...\n");
     return -1;
 
  have_pxenv:
     APIVer = pxenv->version;
     if (!quiet)
-	printf("Found PXENV+ structure\nPXE API version is %04x\n", APIVer);
+	ddprintf("Found PXENV+ structure\nPXE API version is %04x\n", APIVer);
 
     /* if the API version number is 0x0201 or higher, use the !PXE structure */
     if (APIVer >= 0x201) {
@@ -186,6 +221,7 @@ int pxe_init(bool quiet)
     code_seg = pxenv->undicodeseg;
     PXEEntry = pxenv->rmentry;
     type = "PXENV+";
+
     goto have_entrypoint;
 
  have_pxe:
@@ -197,13 +233,18 @@ int pxe_init(bool quiet)
     type = "!PXE";
 
  have_entrypoint:
+    StrucPtr.offs = off;
+    StrucPtr.seg  = seg;
+
     if (!quiet) {
-	printf("%s entry point found (we hope) at %04X:%04X via plan %c\n",
+	ddprintf("%s entry point found (we hope) at %04X:%04X via plan %c\n",
 	       type, PXEEntry.seg, PXEEntry.offs, plan);
-	printf("UNDI code segment at %04X len %04X\n", code_seg, code_len);
-	printf("UNDI data segment at %04X len %04X\n", data_seg, data_len);
+	ddprintf("UNDI code segment at %04X len %04X\n", code_seg, code_len);
+	ddprintf("UNDI data segment at %04X len %04X\n", data_seg, data_len);
     }
 
+    syslinux_memscan_new(pxelinux_scan_memory);
+
     code_seg = code_seg + ((code_len + 15) >> 4);
     data_seg = data_seg + ((data_len + 15) >> 4);
 
@@ -246,7 +287,7 @@ static int pxe_get_cached_info(int type, void *buf, size_t bufsiz)
 {
     int err;
     static __lowmem struct s_PXENV_GET_CACHED_INFO get_cached_info;
-    printf(" %02x", type);
+    ddprintf(" %02x", type);
 
     memset(&get_cached_info, 0, sizeof get_cached_info);
     get_cached_info.PacketType  = type;
@@ -254,7 +295,7 @@ static int pxe_get_cached_info(int type, void *buf, size_t bufsiz)
     get_cached_info.Buffer      = FAR_PTR(buf);
     err = pxe_call(PXENV_GET_CACHED_INFO, &get_cached_info);
     if (err) {
-        printf("PXE API call failed, error  %04x\n", err);
+        ddprintf("PXE API call failed, error  %04x\n", err);
 	kaboom();
     }
 
@@ -312,7 +353,7 @@ __export void unload_pxe(uint16_t flags)
 	memset(&unload_call, 0, sizeof unload_call);
 	err = pxe_call(api, &unload_call);
 	if (err || unload_call.Status != PXENV_STATUS_SUCCESS) {
-	    printf("PXE unload API call %04x failed: 0x%x\n",
+	    ddprintf("PXE unload API call %04x failed: 0x%x\n",
 		   api, unload_call.Status);
 	    goto cant_free;
 	}
@@ -340,11 +381,14 @@ __export void unload_pxe(uint16_t flags)
 	    *(uint32_t *)(4 * 0x1a), int_addr);
 
 cant_free:
-    printf("Failed to free base memory error %04x-%08x (%d/%dK)\n",
+    ddprintf("Failed to free base memory error %04x-%08x (%d/%dK)\n",
 	   api, *(uint32_t *)(4 * 0x1a), bios_fbm(), real_base_mem);
     return;
 }
 
+extern const char bdhcp_data[], adhcp_data[];
+extern const uint32_t bdhcp_len, adhcp_len;
+
 void net_parse_dhcp(void)
 {
     int pkt_len;
@@ -353,18 +397,25 @@ void net_parse_dhcp(void)
 
     bp = lmalloc(dhcp_max_packet);
     if (!bp) {
-	printf("Out of low memory\n");
+	ddprintf("Out of low memory\n");
 	kaboom();
     }
 
     *LocalDomain = 0;   /* No LocalDomain received */
 
+    /*
+     * Parse any "before" hardcoded options
+     */
+    dprintf("DHCP: bdhcp_len = %d\n", bdhcp_len);
+    parse_dhcp_options(bdhcp_data, bdhcp_len, 0);
+
     /*
      * Get the DHCP client identifiers (query info 1)
      */
-    printf("Getting cached packet ");
+    ddprintf("Getting cached packet ");
     pkt_len = pxe_get_cached_info(1, bp, dhcp_max_packet);
     parse_dhcp(bp, pkt_len);
+
     /*
      * We don't use flags from the request packet, so
      * this is a good time to initialize DHCPMagic...
@@ -397,7 +448,13 @@ void net_parse_dhcp(void)
      */
     pkt_len = pxe_get_cached_info(3, bp, dhcp_max_packet);
     parse_dhcp(bp, pkt_len);
-    printf("\n");
+    ddprintf("\n");
+
+    /*
+     * Parse any "after" hardcoded options
+     */
+    dprintf("DHCP: adhcp_len = %d\n", adhcp_len);
+    parse_dhcp_options(adhcp_data, adhcp_len, 0);
 
     lfree(bp);
 }
